#
/*
 *
 *    Copyright (C) 2008, 2009, 2010
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JFF SDR (JSDR).
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	Control object for elektor507. This code is derived from
 *	the Hamlib code together with the Pascal code from
 *	Kainka. All copyrights are gratefully acknowledged.
 */
#include	"elektor507.h"
#include	<stdlib.h>
#include	<string.h>
#include	"jff-include.h"

#define	kHz(f)	((f) * 1000)

	Elektor507::Elektor507	(const char *f, paReader *Reader, bool *success) {
int	res;
	preselection		= f;
	this	-> Reader	= Reader;
	priv			= new priv_data;
	priv	-> xtal_cal	= XTAL_CAL;
	priv	-> osc_freq	= OSCFREQ;
	priv	-> ant		= ANT_AUTO;
//	DIV1N set to safe default
	priv	-> Div1N	= 8;
	priv	-> P		= 8;
	priv	-> Q		= 2;
	rig_is_working		= false;
	ElektorFrequencies	= new frequencytables (preselection);

	if (ftdi_init (&ftdic) < 0) {
	   fprintf (stderr, "ftdi init failed\n");
	   *success = false;
	   return;
	}
//
//	Let us check on the availability of the elektor
//
	if ((res = ftdi_usb_open (&ftdic, 0x0403, 0x6001)) < 0) {
	   fprintf (stderr, "unable to open ftdi %d (%s)\n",
	                    res, ftdi_get_error_string (&ftdic));
	   *success	= false;
	   return;
	}

	ftdi_usb_close	(&ftdic);

	priv	-> FT_port	= 0x03;
	rig_is_working	= true;
	*success	= true;
}

	Elektor507::~Elektor507	(void) {
	ftdi_deinit	(&ftdic);
	delete priv;
	if (Reader != NULL)
	   Reader -> stopReader ();
}

bool	Elektor507::legalFrequency (int32_t f) {
	return 10 < f / 1000 && f / 1000 < 30000;
}

const char	*Elektor507::nameofRig	(void) {
	return "Elektor507";
}

bool	Elektor507::isWorking (void) {
	return rig_is_working;
}

void	Elektor507::setVFOFrequency (int32_t freq) {
int32_t	final_freq;
int	Mux;

	if (!rig_is_working)
	   return;

	if (priv -> ant	== ANT_AUTO) {
	   if (freq < kHz (2000))
	      Mux = 1;
	   else
	      Mux = 2;

	   priv	-> FT_port &= 0x63;	//0, 1 = I2C, 2,3,4 Mux, 5,6 Att
	   priv -> FT_port |= Mux << 2;
	}

	find_P_Q_DIV1N (priv, (int)freq);
	getVFOFrequency (&final_freq);
//	fprintf (stderr,
//	     "freq = %d, final = %d,  delta = %d, Div1 = %d, P = %d, Q = %d\n",
//	                  (int)freq, (int)final_freq,
//	                  (int)(final_freq - freq),
//	                  (int)priv -> Div1N, (int)priv -> P, (int)priv -> Q);

	cy_update_pll (CY_I2C_RAM_ADR);
}

void	Elektor507::setFrequency	(int32_t f1, int32_t f2) {
	(void)f2;
	setVFOFrequency (f1);
}

void	Elektor507::getVFOFrequency (int32_t *freq) {
double	VCO;

	if (!rig_is_working) {
	   *freq = 0;
	   return;
	}

	VCO = ((double)priv -> osc_freq * kHz (1)) / priv -> Q  * priv -> P;
	*freq = (VCO / priv -> Div1N) / 4;
}

void	Elektor507::setLevel (int16_t val) {
int16_t	att;

	if (!rig_is_working)
	   return;

	switch (val) {
	   case 0:	att = 0;	break;
	   case 10:	att = 1;	break;
	   case 20:	att = 2;	break;
	   default:
	      return;
	}

	priv	-> FT_port &= 0x1f;
	priv	-> FT_port |= (att & 0x3) << 5;
	WriteData (&priv -> FT_port, 1);
}

bool	Elektor507::getLevel (int16_t *val) {

	if (!rig_is_working) {
	   *val = 0;
	   return false;
	}

	switch ((priv -> FT_port >> 5) & 0x3) {
	   case 0:	*val = 0;	break;
	   case 1:	*val = 10;	break;
	   case 2:	*val = 20;	break;
	   default:
	      return false;
	}

	return true;
}

bool	Elektor507::setAntenna (int16_t ant) {
uint8_t	Mux;

	if (!rig_is_working)
	   return false;

//	fprintf (stderr, " selecting antenne %d\n", ant);
	switch (ant) {
	   case 1:	Mux = 0;
	                priv -> ant = ANT_EXT;	break;
	   case 2:	
	   case 3:	Mux = 0; 
	                priv -> ant = ANT_AUTO;  break; /* auto		*/
	   case 4:	Mux = 3;
	                priv -> ant = ANT_BAND4; break;	/* 	*/
	   case 5:	Mux = 4;
	                priv -> ant = ANT_BAND5; break;
	   case 6:	Mux = 5;
	                priv -> ant = ANT_BAND6; break;
	   case 7:	Mux = 6;
	                priv -> ant = ANT_BAND7; break;	/* */
	   case 8:	Mux = 7;
	                priv -> ant = ANT_TEST;  break; /* test		*/
	   default:	Mux = 0;
	                priv -> ant = NO_ANT;
	}

	priv	-> FT_port &= 0x63;
	priv	-> FT_port |= Mux << 2;
	return i2c_write_reg (CY_I2C_RAM_ADR,
	                      CLKOE_REG, 
	                      0x20 | (priv -> ant == ANT_TEST ? 0x04 : 0));
}

bool	Elektor507::getAntenna (int16_t *ant) {

	if (!rig_is_working) {
	   *ant = 0;
	   return false;
	}

	*ant = priv -> ant;
	return true;
}

bool	Elektor507::setXtalCal (int16_t n) {

	if (!rig_is_working) 
	   return false;

	priv -> xtal_cal = XTAL_CAL + n;
	return cy_update_pll (CY_I2C_RAM_ADR);
}

bool	Elektor507::preSelect (int32_t f, int16_t *b, int16_t *v) {

	if (!ElektorFrequencies -> findTuningData (f, b, v)) {
	   *b = -1;
	   *v = 10;
	   return false;
	}
	return true;
}

bool	Elektor507::setVoltage (int16_t s) {
	return i2c_write_reg (PCF8951_ADDRESS, DAC_ENABLE, s);
}

bool	Elektor507::cy_update_pll (uint8_t IICaddress) {
int	P0, R40, R41, R42;
uint8_t Div1N;
uint8_t Clk3_src;
int16_t	Pump;
bool	ret;

	Pump	= getPump (priv -> P);
	P0	= priv -> P & 0x01;
	R40	= (((priv -> P >> 1) - 4) >> 8) | (Pump << 2) | 0xc0;
	R41	= (((priv -> P >> 1) - 4) & 0xff);
	R42	= (priv -> Q - 2) | (P0 << 7);

	if (!i2c_write_regs (IICaddress, 3,
	                      PUMPCOUNTERS_REG, R40, R41, R42))
	   return false;

	switch (priv -> Div1N) {
	   case 2:		/* fixed /2 divider option	*/
	      Clk3_src = 0x80;
	      Div1N = 8;
	      break;

	   case 3:		/* fixed /3 divider option	*/
	      Clk3_src = 0xc0;
	      Div1N = 6;
	      break;

	   default:
	      Div1N = priv -> Div1N;
	      Clk3_src = 0x40;
	}

	if (!i2c_write_reg (IICaddress, DIV1_REG, Div1N))
	   return false;

/* set 2 low bits of CLKSRC for clock5, DIV1CLK is already set	*/
	if (!i2c_write_reg (IICaddress, CLKSRC_REG + 2, Clk3_src | 0x07))
	   return false;

	ret = i2c_write_reg (CY_I2C_RAM_ADR, XTALCTL_REG, 0x32);
	ret |= i2c_write_reg (IICaddress, 
	                     CAPLOAD_REG, priv -> xtal_cal);
	return ret;
}

bool	Elektor507::i2c_write_reg (uint8_t IICaddress,
	                           uint8_t reg_addr, uint8_t val) {
	return	i2c_write_regs (IICaddress, 1, reg_addr, val, 0, 0);
}

bool	Elektor507::i2c_write_regs (uint8_t IICaddress, int reg_count,
	                            uint8_t reg_addr, uint8_t reg_val1,
	                            uint8_t reg_val2, uint8_t reg_val3) {

	priv	-> Buf_address = 0;
	ftdi_I2C_Init ();
	ftdi_I2C_Start ();
	ftdi_I2C_Write_Byte (IICaddress);
	ftdi_I2C_Write_Byte (reg_addr);

	if (reg_count >= 1) ftdi_I2C_Write_Byte (reg_val1);
	if (reg_count >= 2) ftdi_I2C_Write_Byte (reg_val2);
	if (reg_count >= 3) ftdi_I2C_Write_Byte (reg_val3);
	ftdi_I2C_Stop ();
//	usleep (10000);
 	return WriteData (priv -> FT_OutBuffer, priv -> Buf_address);
}

void	Elektor507::ftdi_I2C_Write_Byte (uint8_t c) {
int16_t	i;

	for (i = 7; i >= 0; i --) {
	   ftdi_SDA (c & (01 << i));		/* SDA value	*/
	   ftdi_SCL (1);
	   ftdi_SCL (0);
	}

	ftdi_SDA (1); ftdi_SCL (1); ftdi_SCL (0);
}

void	Elektor507::ftdi_I2C_Init () {
	ftdi_SCL (1); ftdi_SDA (1);
}

void	Elektor507::ftdi_I2C_Start () {
	ftdi_SDA (0); ftdi_SCL (0);
}

void	Elektor507::ftdi_I2C_Stop () {
	ftdi_SCL (0); ftdi_SDA (0); ftdi_SCL (1); ftdi_SDA (1);
}

void	Elektor507::ftdi_SDA (int8_t d) {
	if (priv -> Buf_address >= FT_OutBuffer_MAX)
	   return;

	if (d == 0)
	   priv -> FT_port &= ~0x01;
	else
	   priv -> FT_port |= 0x01;

	priv -> FT_OutBuffer [priv -> Buf_address ++] = priv -> FT_port;
}

void	Elektor507::ftdi_SCL (int8_t d) {
	if (priv -> Buf_address >= FT_OutBuffer_MAX)
	   return;

	if (d == 0)
	   priv -> FT_port &= ~0x02;
	else
	   priv -> FT_port |= 0x02;

	priv -> FT_OutBuffer [priv -> Buf_address ++] = priv -> FT_port;
}

void	Elektor507::find_P_Q (struct priv_data *priv, int32_t freq) {
double	Min, VCO;
int	p, q, q_max;
int32_t	freqK	= freq /= kHz (1);
	Min = priv -> osc_freq;
/*
 * 	Q: 2 .. 129
 * 	P: 8 .. 2055, best 16 .. 1023 because of pump
 * 	For stable operation:
 * 	REF /Qtotal must not fall below 250 kHz
 * 	(P * (REF / Qtotal) must not be above 400 Mhz or below 100 Mhz
 */
	q_max = priv -> osc_freq / 250;
	for (q = q_max; q >= 10; q --) 
	   for (p = 500; p <= 2000; p ++) {
	      VCO = ((double) priv -> osc_freq / q) * p;
	      if (fabs (4 * freqK - VCO / priv -> Div1N) < Min) {
	         Min = fabs (4 * freqK - VCO / priv -> Div1N);
	         priv -> Q = q;
	         priv -> P = p;
	      }
	   }

	VCO = ((double)priv -> osc_freq / priv -> Q) * priv -> P;
	if (VCO < 100000 || VCO > 400000)
	   fprintf (stderr, " Unstable for %f, %d %d %d\n",
	            VCO, priv -> P, priv -> Q, priv -> osc_freq);
}

#define	vco_min	100e3
#define	vco_max	500e3

void	Elektor507::find_P_Q_DIV1N (struct priv_data *priv, int32_t f) {
double	Min, VCO, freq4;
int	p, q, div1n, q_max;
int	div1n_min, div1n_max;

	Min	= priv -> osc_freq;
	freq4	= f * 4 / kHz (1);
/*
 * 	Q: 2 .. 129
 * 	P: 8 .. 2055, best 16 .. 1023 (because of pump)
 * 	For stable operations
 * 	REF / Qtotal must not fall below 250 kHz
 * 	P * (REF / Qtotal) must not be above 400 MHz or below 100 Mhz
 */
	q_max = priv -> osc_freq / 250;
	div1n_min	= vco_min / freq4;
	if (div1n_min < 2) 
	   div1n_min = 2;
	if (div1n_min > 127)
	   div1n_min = 127;

	div1n_max  	= vco_max / freq4;
	if (div1n_max > 127)
	   div1n_max = 127;
	if (div1n_max < 2)
	   div1n_max = 2;

	for (div1n = div1n_min; div1n < div1n_max; div1n++) {
// P / Qtotal	= FREQ4 * DIV1N / REF
// (Q * int (r) + frac (r) * Q) / Q`
	   for (q = q_max; q >= 2; q --) {
	      p = q * freq4 * div1n / priv -> osc_freq;
	      if (p < 16 || p > 1023)
	         continue;

	      VCO = ((double) priv -> osc_freq / q) * p;
	      if (VCO < vco_min || VCO > vco_max)
	         continue;

	      if (fabs (freq4 - VCO / div1n) < Min) {
	         Min = fabs (freq4 - VCO / div1n);
	         priv -> Div1N	= div1n;
	         priv -> P	= p;
	         priv -> Q	= q;
	      }
	   }
	}

	VCO = ((double) priv -> osc_freq / priv -> Q) * priv -> P;
	if (VCO < vco_min || VCO > 400e3)
	   fprintf (stderr, "Unstable parameters for %f\n", VCO);
}
	
int16_t	Elektor507::getPump (int16_t p) {
	if (p < 45)
	   return 0;
	if (p < 480)
	   return 1;
	if (p < 640)
	   return 2;
	if (p < 800)
	   return 3;
	return 4;
}

bool	Elektor507::WriteData (void *FT_OutBuf, uint32_t BufferSize) {
int	ret;

	if (!rig_is_working)
	   return false;

	ret = ftdi_usb_open (&ftdic, 0x0403, 0x6001);
	if (ret < 0 && ret != -5) {
	   fprintf (stderr, "unable to open ftdi %d (%s)\n",
	                    ret, ftdi_get_error_string (&ftdic));
	   return false;
	}

	ftdi_set_bitmode (&ftdic, 0xFF, BITMODE_BITBANG);
	ftdi_set_baudrate (&ftdic, 38400 / 4);

 	ret = ftdi_write_data (&ftdic,
	                       (uint8_t *)FT_OutBuf,
	                       BufferSize);
	if (ret < 0) {
	   fprintf (stderr, "writing data did not work %s\n",
	            ftdi_get_error_string (&ftdic));
	   return false;
	}

	ftdi_usb_close	(&ftdic);
	return true;
}

bool	Elektor507::restartReader	(void) {
	if (Reader != NULL)
	   return Reader -> restartReader ();

	return false;
}

void	Elektor507::stopReader	(void) {
	if (Reader != NULL)
	   Reader -> stopReader ();
}

int32_t	Elektor507::getSamples	(DSPCOMPLEX *v, int32_t a, uint8_t m) {
	if (Reader != NULL)
	   return Reader -> getSamples (v, a, m);

	return 0;
}

int32_t	Elektor507::Samples		(void) {
	if (Reader != NULL)
	   return Reader -> Samples ();
	return 0;
}

bool	Elektor507::RigprovidesSamples	(void) {
	return false;
}

